# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function
from webassets.filter import Filter

import dukpy


__all__ = ('TypeScript', )


class TypeScript(Filter):
    name = 'typescript'
    max_debug_level = None

    def input(self, _in, out, **kw):
        src = dukpy.typescript_compile(_in.read())
        out.write(src)



import os
from .evaljs import evaljs

BABEL_COMPILER = os.path.join(os.path.dirname(__file__), 'jsmodules', 'babel-6.14.0.min.js')


def babel_compile(source, **kwargs):
    """Compiles the given ``source`` from ES6 to ES5 usin Babeljs"""
    presets = kwargs.get('presets')
    if not presets:
        kwargs['presets'] = ["es2015"]
    with open(BABEL_COMPILER, 'r') as babel_js:
        return evaljs(
            (babel_js.read(),
             'var bres, res;'
             'bres = Babel.transform(dukpy.es6code, dukpy.babel_options);',
             'res = {map: bres.map, code: bres.code};'),
            es6code=source,
            babel_options=kwargs
        )


def jsx_compile(source, **kwargs):
    kwargs['presets'] = ['es2015', 'react']
    return babel_compile(source, **kwargs)['code']



import os
from .evaljs import evaljs

TS_COMPILER = os.path.join(os.path.dirname(__file__), 'jsmodules', 'typescriptServices.js')
TSC_OPTIONS = '{ module: ts.ModuleKind.System, target: ts.ScriptTarget.ES5, newLine: 1 }'


def typescript_compile(source):
    """Compiles the given ``source`` from TypeScript to ES5 using TypescriptServices.js"""
    with open(TS_COMPILER, 'r') as tsservices_js:
        return evaljs(
            (tsservices_js.read(),
             'ts.transpile(dukpy.tscode, {options});'.format(options=TSC_OPTIONS)),
            tscode=source
        )



import os
from .evaljs import evaljs

COFFEE_COMPILER = os.path.join(os.path.dirname(__file__), 'jsmodules', 'coffeescript.js')


def coffee_compile(source):
    """Compiles the given ``source`` from CoffeeScript to JavaScript"""
    with open(COFFEE_COMPILER, 'r') as coffeescript_js:
        return evaljs(
            (coffeescript_js.read(),
             'CoffeeScript.compile(dukpy.coffeecode)'),
            coffeecode=source
        )



from .nodelike import NodeLikeInterpreter


def less_compile(source, options=None):
    """Compiles the given ``source`` from LESS to CSS"""
    options = options or {}
    res = NodeLikeInterpreter().evaljs(
        ('var result = null;'
         'var less = require("less/less-node");',
         'less.render(dukpy.lesscode, dukpy.lessoptions, function(error, output) {'
         '  result = {"error": error, "output": output};'
         '});'
         'result;'),
        lesscode=source,
        lessoptions=options
    )
    if not res:
        raise RuntimeError('Results or errors unavailable')

    if res.get('error'):
        raise LessCompilerError(res['error']['message'])

    return res['output']['css']


class LessCompilerError(Exception):
    pass




from __future__ import print_function
import json
import os
import logging

from dukpy.module_loader import JSModuleLoader
from . import _dukpy

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

try:  # pragma: no cover
    unicode
    string_types = (str, unicode)
except NameError:  # pragma: no cover
    string_types = (bytes, str)


log = logging.getLogger('dukpy.interpreter')


class JSInterpreter(object):
    """JavaScript Interpreter"""
    def __init__(self):
        self._loader = JSModuleLoader()
        self._ctx = _dukpy.create_context()
        self._funcs = {}

        self._init_process()
        self._init_console()
        self._init_require()

    @property
    def loader(self):
        return self._loader

    def evaljs(self, code, **kwargs):
        """Runs JavaScript code in the context of the interpreter.

        All arguments will be converted to plain javascript objects
        through the JSON encoder and will be available in `dukpy`
        global object.

        Returns the last object on javascript stack.
        """
        jsvars = json.dumps(kwargs)
        jscode = self._adapt_code(code)

        if not isinstance(jscode, bytes):
            jscode = jscode.encode('utf-8')

        if not isinstance(jsvars, bytes):
            jsvars = jsvars.encode('utf-8')

        res = _dukpy.eval_string(self, jscode, jsvars)
        if res is None:
            return None

        return json.loads(res.decode('utf-8'))

    def export_function(self, name, func):
        """Exports a python function to the javascript layer with the given name.

        Note that it is possible to pass back and forth between JS and Python
        only plain javascript objects and that the objects are passed by
        copy so it is not possible to modify a python object from js.
        """
        self._funcs[name] = func

    def _call_python(self, func, json_args):
        # Arguments came in reverse order from JS
        func = func.decode('utf-8')
        json_args = json_args.decode('utf-8')

        args = list(reversed(json.loads(json_args)))
        ret = self._funcs[func](*args)
        if ret is not None:
            return json.dumps(ret).encode('utf-8')

    def _init_process(self):
        self.evaljs("process = {}; process.env = dukpy.environ", environ=dict(os.environ))

    def _init_console(self):
        self.export_function('dukpy.log.info', lambda *args: log.info(' '.join(args)))
        self.export_function('dukpy.log.error', lambda *args: log.error(' '.join(args)))
        self.export_function('dukpy.log.warn', lambda *args: log.warn(' '.join(args)))
        self.evaljs("""
        ;console = {
            log: function() {
                call_python('dukpy.log.info', Array.prototype.join.call(arguments, ' '));
            },
            info: function() {
                call_python('dukpy.log.info', Array.prototype.join.call(arguments, ' '));
            },
            warn: function() {
                call_python('dukpy.log.warn', Array.prototype.join.call(arguments, ' '));
            },
            error: function() {
                call_python('dukpy.log.error', Array.prototype.join.call(arguments, ' '));
            }
        };
        """)

    def _init_require(self):
        self.export_function('dukpy.lookup_module', self._loader.load)
        self.evaljs("""
        ;Duktape.modSearch = function (id, require, exports, module) {
            var m = call_python('dukpy.lookup_module', id);
            if (!m || !m[1]) {
                throw new Error('cannot find module: ' + id);
            }
            _require_set_module_id(require, m[0]);
            return m[1];
        };
""")

    def _adapt_code(self, code):
        def _read_files(f):
            if hasattr(f, 'read'):
                return f.read()
            else:
                return f

        code = _read_files(code)
        if not isinstance(code, string_types) and hasattr(code, '__iter__'):
            code = ';\n'.join(map(_read_files, code))
        return code


def evaljs(code, **kwargs):
    """Evaluates the given ``code`` as JavaScript and returns the result"""
    return JSInterpreter().evaljs(code, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import os

from webassets.filter import Filter, option

import dukpy


__all__ = ('CompileLess', )


class CompileLess(Filter):
    name = 'lessc'
    max_debug_level = None
    options = {
        'less_includes': option('LIBSASS_INCLUDES', type=list)
    }

    def input(self, _in, out, **kw):
        options = {'paths': []}
        if self.less_includes:
            options['paths'].extend(self.less_includes)
        if 'source_path' in kw:
            options['paths'].append(os.path.dirname(kw['source_path']))

        src = dukpy.less_compile(_in.read(), options=options)
        out.write(src)



# -*- coding: utf-8 -*-
import json
import os


class JSModuleLoader(object):
    """Manages finding and loading JS modules in CommonJS format.

    This allows to import a module from JSInterpreter using the
    `require('modulename')` command.

    To register additional paths where to look for modules use
    the `JSModuleLoader.register_path` method.
    """
    def __init__(self):
        self._paths = []
        self.register_path(os.path.join(os.path.dirname(__file__), 'jsmodules'))
        self.register_path(os.getcwd())

    def register_path(self, path):
        """Registers a directory where to look for modules.

        By default only modules relative to current path are found.
        """
        self._paths.insert(0, os.path.abspath(path))

    def lookup(self, module_name):
        """Searches for a file providing given module.

        Returns the normalized module id and path of the file.
        """
        for search_path in self._paths:
            module_path = os.path.join(search_path, module_name)
            new_module_name, module_file = self._lookup(module_path, module_name)
            if module_file:
                return new_module_name, module_file
        return None, None

    def load(self, module_name):
        """Returns source code and normalized module id of the given module.

        Only supports source code files encoded as UTF-8
        """
        module_name, path = self.lookup(module_name)
        if path:
            with open(path, 'rb') as f:
                return module_name, f.read().decode('utf-8')
        return None, None

    def _lookup(self, module_path, module_name):
        # Module is a plain .js file
        for path in (module_path, os.path.extsep.join((module_path, 'js'))):
            if os.path.exists(path) and os.path.isfile(path):
                return module_name, path

        # Module is a package
        package = os.path.join(module_path, os.path.extsep.join(('package', 'json')))
        try:
            with open(package) as f:
                package = json.load(f)
        except IOError:
            pass
        else:
            package_main = package.get('main')
            if package_main:
                main_name = package_main
                if package_main.endswith('.js'):
                    main_name = package_main[:-3]
                main_name, path = self._lookup(os.path.join(module_path, package_main), main_name)
                if path:
                    return '/'.join((module_name, main_name)), path

        # Module is directory with index.js inside
        indexjs = os.path.join(module_path, os.path.extsep.join(('index', 'js')))
        if os.path.exists(indexjs):
            return '/'.join((module_name, 'index')), indexjs

        return None, None



# -*- coding: utf-8 -*-
from __future__ import print_function
import argparse
import json
import os
import sys
import tarfile
import tempfile
import shutil
from contextlib import closing
from io import BytesIO
from .evaljs import evaljs

try:
    from urllib.request import urlopen
    from urllib.parse import quote_plus
except ImportError:
    from urllib2 import urlopen
    from urllib import quote_plus


def main():
    parser = argparse.ArgumentParser(description='Install a Javascript Package from npmjs.org')
    parser.add_argument('package_name',
                        help='name of the package to install')
    parser.add_argument('version', nargs='?',
                        help='Version of the package (omit for latest)')
    parser.add_argument('--destination', '-d', default='./js_modules',
                        help="directory where to install javascript packages")
    args = parser.parse_args(sys.argv[1:])

    try:
        return install_jspackage(args.package_name, args.version, args.destination)
    except JSPackageInstallError as e:
        print(e)
        return e.error_code


def install_jspackage(package_name, version, modulesdir):
    """Installs a JavaScript package downloaded from npmjs.org.

    For example to install React::

        install_jspackage('react', '0.14.8', './node_modules')

    To install last version provide `None` as the version.
    """
    if not version:
        version = ''

    requirements = _resolve_dependencies(package_name, version)
    print('Packages going to be installed: {0}'.format(', '.join(
        '{0}->{1}'.format(*i) for i in requirements
    )))

    downloads = {}
    for dependency_name, _, version_info in requirements:
        try:
            downloads[dependency_name] = version_info['dist']['tarball']
        except KeyError:
            raise JSPackageInstallError('Unable to detect a supported download url for package',
                                        error_code=3)

    for dependency_name, download_url in downloads.items():
        tarball = BytesIO()
        print('Fetching {0}'.format(download_url), end='')
        with closing(urlopen(download_url)) as data:
            chunk = data.read(1024)
            while chunk:
                print('.', end='')
                tarball.write(chunk)
                chunk = data.read(1024)
        print('')

        tarball.seek(0)
        with closing(tarfile.open(fileobj=tarball)) as tb:
            dest = os.path.join(modulesdir, dependency_name)
            tmpdir = tempfile.mkdtemp()
            try:
                tb.extractall(tmpdir)
                shutil.rmtree(os.path.abspath(dest), ignore_errors=True)
                shutil.move(os.path.join(tmpdir, 'package'),
                            os.path.abspath(dest))
            finally:
                shutil.rmtree(tmpdir)

    print('Installing {0} in {1} Done!'.format(package_name, modulesdir))


def _resolve_version(version, versions):
    return evaljs('''
        var semver = require('semver');
        semver.maxSatisfying(dukpy.versions, dukpy.version)
    ''', version=version, versions=list(versions.keys()))


def _fetch_package_info(package_name):
    url = 'http://registry.npmjs.org/{0}'
    with closing(urlopen(url.format(package_name))) as data:
        return json.loads(data.read().decode('utf-8'))


def _resolve_dependencies(package_name, version):
    package_info = _fetch_package_info(quote_plus(package_name, safe='@'))
    package_versions = package_info['versions']
    matching_version = _resolve_version(version, package_versions)
    version_info = package_versions.get(matching_version)
    if version_info is None:
        raise JSPackageInstallError('Version {0} not found, available versions are {1}'.format(
            version, ', '.join(sorted(package_versions.keys()))
        ), error_code=2)

    requirements = [(package_name, matching_version, version_info)]
    dependencies = version_info.get('dependencies', {})
    if dependencies:
        for dependency, dependency_version in dependencies.items():
            requirements.extend(
                _resolve_dependencies(dependency, dependency_version)
            )
    return requirements


class JSPackageInstallError(Exception):
    def __init__(self, msg, error_code):
        super(JSPackageInstallError, self).__init__(msg)
        self.error_code = error_code



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import os

from webassets.filter import Filter

import dukpy


__all__ = ('BabelJS', )


class BabelJS(Filter):
    name = 'babeljs'
    max_debug_level = None
    options = {
        'loader': 'BABEL_MODULES_LOADER'
    }

    def input(self, _in, out, **kw):
        options = {'filename': os.path.basename(kw['source_path'])}
        if self.loader == 'systemjs':
            options['plugins'] = ['transform-es2015-modules-systemjs']
        elif self.loader == 'umd':
            options['plugins'] = ['transform-es2015-modules-umd']
        src = dukpy.babel_compile(_in.read(), **options)
        out.write(src['code'])



from .evaljs import evaljs, JSInterpreter
from ._dukpy import JSRuntimeError
from .install import install_jspackage

from .coffee import coffee_compile
from .babel import babel_compile, jsx_compile
from .tsc import typescript_compile
from .lessc import less_compile


# -*- coding: utf-8 -*-
import argparse
import sys
from .nodelike import NodeLikeInterpreter


def main():
    parser = argparse.ArgumentParser(description='Run a javascript script')
    parser.add_argument('filename', help='path of the script to run')
    args = parser.parse_args(sys.argv[1:])

    with open(args.filename) as f:
        sourcecode = f.read()

        if sourcecode.startswith('#!'):
            # Remove shebang
            _, sourcecode = sourcecode.split('\n', 1)
            sourcecode = '\n' + sourcecode

        runner = NodeLikeInterpreter()
        runner.evaljs(sourcecode)


#!/usr/bin/env python
import os
from distutils.core import setup, Extension

import sys
py_version = sys.version_info[:2]

HERE = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(HERE, 'README.rst')).read()
except IOError:
    README = ''

INSTALL_REQUIRES = []
if py_version == (2, 6):
    INSTALL_REQUIRES.append('argparse')

duktape = Extension('dukpy._dukpy',
                    define_macros=[('DUK_OPT_DEEP_C_STACK', '1'),
                                   ('DUK_OPT_NONSTD_REGEXP_DOLLAR_ESCAPE', '1'),
                                   ('DUK_OPT_OCTAL_SUPPORT', '1')],
                    sources=[os.path.join('src', 'duktape', 'duktape.c'),
                             os.path.join('src','_support.c'),
                             os.path.join('src','pyduktape.c')],
                    include_dirs=[os.path.join('.', 'src', 'duktape')])

setup(
    name='dukpy',
    version='0.1.0',
    description='Simple JavaScript interpreter for Python',
    long_description=README,
    keywords='javascript compiler babeljs jsx coffeescript typescript',
    author='Alessandro Molina',
    author_email='alessandro.molina@axant.it',
    url='https://github.com/amol-/dukpy',
    license='MIT',
    packages=['dukpy', 'dukpy.webassets'],
    ext_modules=[duktape],
    install_requires=INSTALL_REQUIRES,
    package_data={
        'dukpy': ['jscore/*.js', 'jsmodules/*.js', 'jsmodules/react/*.js',
                  'jsmodules/less/*/*.js', 'jsmodules/less/*/*/*.js'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: JavaScript',
    ],
    entry_points={
        'console_scripts': [
            'dukpy-install = dukpy.install:main',
            'dukpy = dukpy.run:main'
        ]
    }
)



# -*- coding: utf-8 -*-
import os
from .evaljs import JSInterpreter


class NodeLikeInterpreter(JSInterpreter):
    """A DukPy Interpreter that provides a minimal compatibility layer with NodeJS"""
    def __init__(self):
        super(NodeLikeInterpreter, self).__init__()
        self.loader.register_path(os.path.join(os.path.dirname(__file__), 'jscore'))
        self.export_function('file.exists', FS.exists)
        self.export_function('file.read', FS.read)


class FS:
    """Provides oversimplified fs.js native functions."""
    @classmethod
    def exists(cls, filepath):
        try:
            os.stat(filepath)
            return True
        except:
            return False

    @classmethod
    def read(cls, path, encoding):
        with open(path, 'rb') as f:
            data = f.read()

        if encoding is not None:
            return data.decode(encoding)
        else:
            return data
